@@ -3,22 +3,22 @@
 """
 Query Debian's Bug Tracking System (BTS).
 
-This module provides a layer between Python and Debian's BTS. It
-provides methods to query the BTS using the BTS' SOAP interface, and the
-Bugreport class which represents a bugreport from the BTS.
+This module provides a layer between Python and Debian's BTS. It provides
+methods to query the BTS using the BTS' SOAP interface, and the Bugreport class
+which represents a bugreport from the BTS.
 """
 
 
+from __future__ import annotations
+
 import base64
-from distutils.version import LooseVersion
 import email.feedparser
 import email.policy
 from datetime import datetime
 import os
-import sys
 import logging
+from typing import Any, Iterable
 
-import pysimplesoap
 from pysimplesoap.client import SoapClient
 from pysimplesoap.simplexml import SimpleXMLElement
 
@@ -32,9 +32,6 @@ if os.path.isdir(ca_path):
     os.environ["SSL_CERT_DIR"] = ca_path
 
 
-PYSIMPLESOAP_1_16_2 = (LooseVersion(pysimplesoap.__version__) >=
-                       LooseVersion('1.16.2'))
-
 # Setup the soap server
 # Default values
 URL = "https://bugs.debian.org/cgi-bin/soap.cgi"
@@ -54,27 +51,25 @@ SEVERITIES = {
 }
 
 
-class Bugreport(object):
+class Bugreport:
     """Represents a bugreport from Debian's Bug Tracking System.
 
-    A bugreport object provides all attributes provided by the SOAP
-    interface. Most of the attributes are strings, the others are
-    marked.
+    A bugreport object provides all attributes provided by the SOAP interface.
+    Most of the attributes are strings, the others are marked.
 
     Attributes
     ----------
-
     bug_num : int
         The bugnumber
     severity : str
         Severity of the bugreport
-    tags : list of strings
+    tags : list[str]
         Tags of the bugreport
     subject : str
         The subject/title of the bugreport
     originator : str
         Submitter of the bugreport
-    mergedwith : list of ints
+    mergedwith : list[int]
         List of bugnumbers this bug was merged with
     package : str
         Package of the bugreport
@@ -86,21 +81,21 @@ class Bugreport(object):
         Date of update of the bugreport
     done : boolean
         Is the bug fixed or not
-    done_by : str or None
+    done_by : str | None
         Name and Email or None
     archived : bool
         Is the bug archived or not
     unarchived : bool
         Was the bug unarchived or not
-    fixed_versions : list of strings
+    fixed_versions : list[str]
         List of versions, can be empty even if bug is fixed
-    found_versions : list of strings
+    found_versions : list[str]
         List of version numbers where bug was found
     forwarded : str
         A URL or email address
-    blocks: list of ints
+    blocks: list[int]
         List of bugnumbers this bug blocks
-    blockedby : list of int
+    blockedby : list[int]
         List of bugnumbers which block this bug
     pending : str
         Either 'pending' or 'done'
@@ -110,37 +105,38 @@ class Bugreport(object):
         Who took responsibility for fixing this bug
     location : str
         Either 'db-h' or 'archive'
-    affects : list of str
+    affects : list[str]
         List of Packagenames
     summary : str
         Arbitrary text
     """
 
-    def __init__(self):
-        self.originator = None
-        self.date = None
-        self.subject = None
-        self.msgid = None
-        self.package = None
-        self.tags = None
-        self.done = None
-        self.forwarded = None
-        self.mergedwith = None
-        self.severity = None
-        self.owner = None
-        self.found_versions = None
-        self.fixed_versions = None
-        self.blocks = None
-        self.blockedby = None
-        self.unarchived = None
-        self.summary = None
-        self.affects = None
-        self.log_modified = None
-        self.location = None
-        self.archived = None
-        self.bug_num = None
-        self.source = None
-        self.pending = None
+    def __init__(self) -> None:
+        self.originator: str
+        self.date: datetime
+        self.subject: str
+        self.msgid: str
+        self.package: str
+        self.tags: list[str]
+        self.done: bool
+        self.done_by: str | None
+        self.forwarded: str
+        self.mergedwith: list[int]
+        self.severity: str
+        self.owner: str
+        self.found_versions: list[str]
+        self.fixed_versions: list[str]
+        self.blocks: list[int]
+        self.blockedby: list[int]
+        self.unarchived: bool
+        self.summary: str
+        self.affects: list[str]
+        self.log_modified: datetime
+        self.location: str
+        self.archived: bool
+        self.bug_num: int
+        self.source: str
+        self.pending: str
         # The ones below are also there but not used
         # self.fixed = None
         # self.found = None
@@ -149,14 +145,13 @@ class Bugreport(object):
         # self.keywords = None
         # self.id = None
 
-    def __str__(self):
+    def __str__(self) -> str:
         s = "\n".join(
-            "{}: {}".format(key, value)
-            for key, value in self.__dict__.items()
+            f"{key}: {value}" for key, value in self.__dict__.items()
         )
         return s + "\n"
 
-    def __lt__(self, other):
+    def __lt__(self, other: Bugreport) -> bool:
         """Compare a bugreport with another.
 
         The more open and urgent a bug is, the greater the bug is:
@@ -165,31 +160,35 @@ class Bugreport(object):
 
             critical > grave > serious > important > normal > minor > wishlist.
 
-        Openness always beats urgency, eg an archived bug is *always*
-        smaller than an outstanding bug.
+        Openness always beats urgency, eg an archived bug is *always* smaller
+        than an outstanding bug.
 
-        This sorting is useful for displaying bugreports in a list and
-        sorting them in a useful way.
+        This sorting is useful for displaying bugreports in a list and sorting
+        them in a useful way.
 
         """
         return self._get_value() < other._get_value()
 
-    def __le__(self, other):
+    def __le__(self, other: Bugreport) -> bool:
         return not self.__gt__(other)
 
-    def __gt__(self, other):
+    def __gt__(self, other: Bugreport) -> bool:
         return self._get_value() > other._get_value()
 
-    def __ge__(self, other):
+    def __ge__(self, other: Bugreport) -> bool:
         return not self.__lt__(other)
 
-    def __eq__(self, other):
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, Bugreport):
+            return NotImplemented
         return self._get_value() == other._get_value()
 
-    def __ne__(self, other):
+    def __ne__(self, other: object) -> bool:
+        if not isinstance(other, Bugreport):
+            return NotImplemented
         return not self.__eq__(other)
 
-    def _get_value(self):
+    def _get_value(self) -> int:
         if self.archived:
             # archived and done
             val = 0
@@ -203,42 +202,37 @@ class Bugreport(object):
         return val
 
 
-def get_status(nrs, *additional):
+def get_status(
+    nrs: int | list[int] | tuple[int, ...],
+) -> list[Bugreport]:
     """Returns a list of Bugreport objects.
 
-    Given a list of bugnumbers this method returns a list of Bugreport
+    Given a list of bug numbers this method returns a list of Bugreport
     objects.
 
     Parameters
     ----------
-    nrs : int or list of ints
+    nrs
         The bugnumbers
-    additional : int
-        Deprecated! The remaining positional arguments are treated as
-        bugnumbers. This is deprecated since 2.10.0, please use the
-        `nrs` parameter instead.
 
     Returns
     -------
-    bugs : list of Bugreport objects
+    list[Bugreport]
+        list of Bugreport objects
 
     """
+    numbers: list[int]
     if not isinstance(nrs, (list, tuple)):
-        nrs = [nrs]
-    # backward compatible with <= 2.10.0
-    if additional:
-        logger.warning(
-            "Calling get_status with bugnumbers as positional"
-            " arguments is deprecated, please use a list instead."
-        )
-        nrs.extend(additional)
+        numbers = [nrs]
+    else:
+        numbers = list(nrs)
 
     # Process the input in batches to avoid hitting resource limits on
     # the BTS
     soap_client = _build_soap_client()
     bugs = []
-    for i in range(0, len(nrs), BATCH_SIZE):
-        slice_ = nrs[i:i + BATCH_SIZE]
+    for i in range(0, len(numbers), BATCH_SIZE):
+        slice_ = numbers[i:i + BATCH_SIZE]
         # I build body by hand, pysimplesoap doesn't generate soap Arrays
         # without using wsdl
         method_el = SimpleXMLElement("<get_status></get_status>")
@@ -250,37 +244,27 @@ def get_status(nrs, *additional):
     return bugs
 
 
-def get_usertag(email, tags=None, *moretags):
+def get_usertag(
+    email: str,
+    tags: None | list[str] | tuple[str, ...] = None,
+) -> dict[str, list[int]]:
     """Get buglists by usertags.
 
     Parameters
     ----------
-    email : str
-    tags : list of strings
-        If tags are given the dictionary is limited to the matching
-        tags, if no tags are given all available tags are returned.
-    moretags : str
-        Deprecated! The remaining positional arguments are treated as
-        tags. This is deprecated since 2.10.0, please use the `tags`
-        parameter instead.
+    email
+    tags
+        If tags are given the dictionary is limited to the matching tags, if no
+        tags are given all available tags are returned.
 
     Returns
     -------
-    mapping : dict
+    dict[str, list[int]]
         a mapping of usertag -> buglist
 
     """
     if tags is None:
         tags = []
-    # backward compatible with <= 2.10.0
-    if not isinstance(tags, (list, tuple)):
-        tags = [tags]
-    if moretags:
-        logger.warning(
-            "Calling get_getusertag with tags as positional"
-            " arguments is deprecated, please use a list instead."
-        )
-        tags.extend(moretags)
 
     reply = _soap_client_call("get_usertag", email, *tags)
     map_el = reply("s-gensym3")
@@ -302,7 +286,9 @@ def get_usertag(email, tags=None, *moretags):
     return mapping
 
 
-def get_bug_log(nr):
+def get_bug_log(
+    nr: int,
+) -> list[dict[str, str | list[Any] | int | email.message.Message]]:
     """Get Buglogs.
 
     A buglog is a dictionary with the following mappings:
@@ -314,51 +300,60 @@ def get_bug_log(nr):
 
     Parameters
     ----------
-    nr : int
+    nr
         the bugnumber
 
     Returns
     -------
-    buglogs : list of dicts
+    list[dict[str, str | list[Any] | int | email.message.Message]]
+        list of buglogs
 
     """
     reply = _soap_client_call("get_bug_log", nr)
     items_el = reply("soapenc:Array")
     buglogs = []
     for buglog_el in items_el.children():
-        buglog = {}
-        buglog["header"] = _parse_string_el(buglog_el("header"))
-        buglog["body"] = _parse_string_el(buglog_el("body"))
-        buglog["msg_num"] = int(buglog_el("msg_num"))
+        buglog: dict[str, str | list[Any] | int | email.message.Message] = {}
+        header = _parse_string_el(buglog_el("header"))
+        body = _parse_string_el(buglog_el("body"))
+        msg_num = int(buglog_el("msg_num"))
         # server always returns an empty attachments array ?
-        buglog["attachments"] = []
+        attachments: list[Any] = []
 
         mail_parser = email.feedparser.BytesFeedParser(
             policy=email.policy.SMTP
         )
-        mail_parser.feed(buglog["header"].encode())
-        mail_parser.feed("\n\n".encode())
-        mail_parser.feed(buglog["body"].encode())
-        buglog["message"] = mail_parser.close()
+        mail_parser.feed(header.encode())
+        mail_parser.feed(b"\n\n")
+        mail_parser.feed(body.encode())
+        message = mail_parser.close()
+
+        buglog = {
+            "header": header,
+            "body": body,
+            "msg_num": msg_num,
+            "attachments": attachments,
+            "message": message,
+        }
 
         buglogs.append(buglog)
     return buglogs
 
 
-def newest_bugs(amount):
+def newest_bugs(amount: int) -> list[int]:
     """Returns the newest bugs.
 
     This method can be used to query the BTS for the n newest bugs.
 
     Parameters
     ----------
-    amount : int
-        the number of desired bugs. E.g. if `amount` is 10 the method
-        will return the 10 latest bugs.
+    amount
+        the number of desired bugs. E.g. if `amount` is 10 the method will
+        return the 10 latest bugs.
 
     Returns
     -------
-    bugs : list of int
+    list[int]
         the bugnumbers
 
     """
@@ -367,18 +362,16 @@ def newest_bugs(amount):
     return [int(item_el) for item_el in items_el.children() or []]
 
 
-def get_bugs(*key_value, **kwargs):
+def get_bugs(
+    **kwargs: str | int | list[int],
+) -> list[int]:
     """Get list of bugs matching certain criteria.
 
     The conditions are defined by the keyword arguments.
 
     Arguments
     ---------
-    key_value : str
-        Deprecated! The positional arguments are treated as key-values.
-        This is deprecated since 2.10.0, please use the `kwargs`
-        parameters instead.
-    kwargs :
+    kwargs
         Possible keywords are:
             * "package": bugs for the given package
             * "submitter": bugs from the submitter
@@ -387,24 +380,24 @@ def get_bugs(*key_value, **kwargs):
             * "severity": bugs with a certain severity
             * "status": can be either "done", "forwarded", or "open"
             * "tag": see http://www.debian.org/Bugs/Developer#tags for
-               available tags
+              available tags
             * "owner": bugs which are assigned to `owner`
             * "bugs": takes single int or list of bugnumbers, filters the list
-               according to given criteria
+              according to given criteria
             * "correspondent": bugs where `correspondent` has sent a mail to
-            * "archive": takes a string: "0" (unarchived), "1"
-              (archived) or "both" (un- and archived). if omitted, only
-              returns un-archived bugs.
+            * "archive": takes a string: "0" (unarchived), "1" (archived) or
+              "both" (un- and archived). if omitted, only returns un-archived
+              bugs.
 
     Returns
     -------
-    bugs : list of ints
+    list[int]
         the bugnumbers
 
     Examples
     --------
-    >>> get_bugs(package='gtk-qt-engine', severity='normal')
-    [12345, 23456]
+        >>> get_bugs(package='gtk-qt-engine', severity='normal')
+        [12345, 23456]
 
     """
     # flatten kwargs to list:
@@ -413,20 +406,6 @@ def get_bugs(*key_value, **kwargs):
     for k, v in kwargs.items():
         args.extend([k, v])
 
-    # previous versions also accepted
-    # get_bugs(['package', 'gtk-qt-engine', 'severity', 'normal'])
-    # if key_value is a list in a one elemented tuple, remove the
-    # wrapping list
-    if len(key_value) == 1 and isinstance(key_value[0], list):
-        key_value = tuple(key_value[0])
-
-    if key_value:
-        logger.warning(
-            "Calling get_bugs with positional arguments is"
-            " deprecated, please use keyword arguments instead."
-        )
-        args.extend(key_value)
-
     # pysimplesoap doesn't generate soap Arrays without using wsdl
     # I build body by hand, converting list to array and using standard
     # pysimplesoap marshalling for other types
@@ -444,8 +423,20 @@ def get_bugs(*key_value, **kwargs):
     return [int(item_el) for item_el in items_el.children() or []]
 
 
-def _parse_status(bug_el):
-    """Return a bugreport object from a given status xml element"""
+def _parse_status(bug_el: SimpleXMLElement) -> Bugreport:
+    """Return a bugreport object from a given status xml element
+
+    Parameters
+    ----------
+    bug_el
+        a status XML element
+
+    Returns
+    -------
+    Bugreport
+        a Bugreport object
+
+    """
     bug = Bugreport()
 
     # plain fields
@@ -465,9 +456,7 @@ def _parse_status(bug_el):
         setattr(bug, field, _parse_string_el(bug_el(field)))
 
     bug.date = datetime.utcfromtimestamp(float(bug_el("date")))
-    bug.log_modified = datetime.utcfromtimestamp(
-        float(bug_el("log_modified"))
-    )
+    bug.log_modified = datetime.utcfromtimestamp(float(bug_el("log_modified")))
     bug.tags = [tag for tag in str(bug_el("tags")).split()]
     bug.done = _parse_bool(bug_el("done"))
     bug.done_by = _parse_string_el(bug_el("done")) if bug.done else None
@@ -498,58 +487,54 @@ def _parse_status(bug_el):
     return bug
 
 
-# to support python 3.4.3, when using httplib2 as pysimplesoap transport
-# we must work around a bug in httplib2, which uses
-# http.client.HTTPSConnection with check_hostname=True, but with an
-# empty ssl context that prevents the certificate verification. Passing
-# `cacert` to httplib2 through pysimplesoap permits to create a valid
-# ssl context.
 _soap_client_kwargs = {
     "location": URL,
     "action": "",
     "namespace": NS,
     "soap_ns": "soap",
 }
-if sys.version_info.major == 3 and sys.version_info < (3, 4, 3):
-    try:
-        from httplib2 import CA_CERTS
-    except ImportError:
-        pass
-    else:
-        _soap_client_kwargs["cacert"] = CA_CERTS
 
 
-def set_soap_proxy(proxy_arg):
+def set_soap_proxy(proxy_arg: str) -> None:
     """Set proxy for SOAP client.
 
     You must use this method after import to set the proxy.
 
     Parameters
     ----------
-    proxy_arg : str
+    proxy_arg
 
     """
     _soap_client_kwargs["proxy"] = proxy_arg
 
 
-def set_soap_location(url):
+def set_soap_location(url: str) -> None:
     """Set location URL for SOAP client
 
     You may use this method after import to override the default URL.
 
     Parameters
     ----------
-    url : str
+    url
+        default URL
 
     """
     _soap_client_kwargs["location"] = url
 
 
-def get_soap_client_kwargs():
+def get_soap_client_kwargs() -> dict[str, str]:
+    """Returns SOAP client kwargs.
+
+    Returns
+    -------
+    dict[str, str]
+        the SOAP client kwargs
+
+    """
     return _soap_client_kwargs
 
 
-def _build_soap_client():
+def _build_soap_client() -> SoapClient:
     """Factory method that creates a SoapClient.
 
     For thread-safety we create SoapClients on demand instead of using a
@@ -557,18 +542,32 @@ def _build_soap_client():
 
     Returns
     -------
-    sc : SoapClient instance
+    SoapClient
+        a SoapClient instance
 
     """
     return SoapClient(**_soap_client_kwargs)
 
 
-def _convert_soap_method_args(*args):
+def _convert_soap_method_args(*args: Iterable[Any]) -> list[tuple[str, Any]]:
     """Convert arguments to be consumed by a SoapClient method
 
+    Parameters
+    ----------
+    *args
+        any argument
+
+    Returns
+    -------
+    list[tuple[str, Any]]
+        the converted arguments
+
+    Examples
+    --------
     Soap client required a list of named arguments:
-    >>> _convert_soap_method_args('a', 1)
-    [('arg0', 'a'), ('arg1', 1)]
+
+        >>> _convert_soap_method_args('a', 1)
+        [('arg0', 'a'), ('arg1', 1)]
 
     """
     soap_args = []
@@ -577,45 +576,94 @@ def _convert_soap_method_args(*args):
     return soap_args
 
 
-def _soap_client_call(method_name, *args):
-    """Wrapper to call SoapClient method"""
+def _soap_client_call(method_name: str, *args: Any) -> Any:
+    """Wrapper to call SoapClient method
+
+    Parameters
+    ----------
+    method_name
+        the method name
+    *args
+
+    Returns
+    -------
+    Any
+
+    """
     # a new client instance is built for threading issues
     soap_client = _build_soap_client()
     soap_args = _convert_soap_method_args(*args)
-    # if pysimplesoap version requires it, apply a workaround for
-    # https://github.com/pysimplesoap/pysimplesoap/issues/31
-    if PYSIMPLESOAP_1_16_2:
-        return getattr(soap_client, method_name)(*soap_args)
-    else:
-        return getattr(soap_client, method_name)(soap_client, *soap_args)
+    return getattr(soap_client, method_name)(*soap_args)
 
 
-def _build_int_array_el(el_name, parent, list_):
-    """build a soapenc:Array made of ints called `el_name` as a child
-    of `parent`"""
+def _build_int_array_el(
+    el_name: str,
+    parent: SimpleXMLElement,
+    list_: list[Any],
+) -> SimpleXMLElement:
+    """Build Array as child of parent.
+
+    More specifically: Build a soapenc:Array made of ints called `el_name` as a
+    child of `parent`.
+
+    Parameters
+    ----------
+    el_name
+    parent
+    list
+
+    Returns
+    -------
+    SimpleXMLElement
+
+    """
     el = parent.add_child(el_name)
     el.add_attribute(
         "xmlns:soapenc", "http://schemas.xmlsoap.org/soap/encoding/"
     )
     el.add_attribute("xsi:type", "soapenc:Array")
-    el.add_attribute("soapenc:arrayType", "xsd:int[{:d}]".format(len(list_)))
+    el.add_attribute("soapenc:arrayType", f"xsd:int[{len(list_):d}]")
     for item in list_:
         item_el = el.add_child("item", str(item))
         item_el.add_attribute("xsi:type", "xsd:int")
     return el
 
 
-def _parse_bool(el):
-    """parse a boolean value from a xml element"""
+def _parse_bool(el: SimpleXMLElement) -> bool:
+    """Parse a boolean value from a XML element.
+
+    Parameters
+    ----------
+    el
+        the element to parse
+
+    Returns
+    -------
+    bool
+        the parsed value
+
+    """
     value = str(el)
     return not value.strip() in ("", "0")
 
 
-def _parse_string_el(el):
-    """read a string element, maybe encoded in base64"""
+def _parse_string_el(el: SimpleXMLElement) -> str:
+    """Read a string element, maybe encoded in base64.
+
+    Parameters
+    ----------
+    el
+        the element to parse
+
+    Returns
+    -------
+    str
+        the parsed value
+
+    """
     value = str(el)
     el_type = el.attributes().get("xsi:type")
     if el_type and el_type.value == "xsd:base64Binary":
-        value = base64.b64decode(value)
-        value = value.decode("utf-8", errors="replace")
+        tmp = base64.b64decode(value)
+        value = tmp.decode("utf-8", errors="replace")
     return value
